mod data;
mod generators;
mod templates;
mod parser;

use std::fmt::Write;
use std::fs::File;
use std::io::Read;
use clap::Parser;

#[derive(Debug, clap::Parser)]
struct Args {
    pub file: std::path::PathBuf,
    #[arg(short='n', long="name", help="base name for output files", required=true)]
    pub rpc_name: std::path::PathBuf,
    #[command(flatten)]
    pub generators: GenArgs
}

#[derive(Debug, clap::Args)]
#[group(required = true, multiple = true)]
struct GenArgs {
    #[arg(value_enum, short='c', long="client")]
    pub clients: Vec<generators::ClientGenerators>,
    #[arg(value_enum, short='s', long="server")]
    pub servers: Vec<generators::ServerGenerators>
}

static SOURCE_FILE: once_cell::sync::OnceCell<String> = once_cell::sync::OnceCell::new();
static SOURCE: once_cell::sync::OnceCell<String> = once_cell::sync::OnceCell::new();

fn main() {
    let args = Args::parse();

    let mut file = File::open(args.file.clone()).unwrap();
    let mut content = String::new();
    file.read_to_string(&mut content).unwrap();

    SOURCE_FILE.set(args.file.to_string_lossy().to_string()).unwrap();
    SOURCE.set(content).unwrap();

    let ast = match syn::parse_file(SOURCE.get().unwrap()) {
        Ok(v) => v,
        Err(e) => emit_error(e.into_iter()
            .map(|e| {
                let span = e.span();
                let mut msg = String::new();
                write!(msg, "{e}").unwrap();
                (span, msg)
            })
        )
    };

    let rpc = parser::parse_file(&ast);

    for gen in &args.generators.clients {
        gen.generate(&args.rpc_name, &rpc);
    }

    for gen in &args.generators.servers {
        gen.generate(&args.rpc_name, &rpc);
    }
}

pub fn emit_error(errors: impl IntoIterator<Item=(proc_macro2::Span, impl Into<String>)>) -> ! {
    use codespan_reporting::{
        diagnostic::{Diagnostic, Label},
        files::{SimpleFiles, Files},
        term::{
            self,
            termcolor::{ColorChoice, StandardStream},
        },
    };
    let mut files = SimpleFiles::new();
    let config = term::Config::default();
    let writer = StandardStream::stderr(ColorChoice::Auto);
    let file_id = files.add(SOURCE_FILE.get().unwrap().clone(), SOURCE.get().unwrap().clone());

    let file = files.get(file_id).unwrap();

    let errors = errors.into_iter().map(|(span, msg)| {
        let start = span.start();
        let start: usize = file.line_range((), start.line-1).unwrap().start + start.column;
        let end = span.end();
        let end: usize = file.line_range((), end.line-1).unwrap().start + end.column;
        Label::primary(file_id, start..end).with_message(msg)
    });

    let diagnostic = Diagnostic::error()
        .with_labels(errors.collect());

    term::emit(&mut writer.lock(), &config, &files, &diagnostic).expect("cannot write error");
    std::process::exit(1);
}
