use syn::spanned::Spanned;
use super::{data, emit_error};

fn parse_enum(item: &syn::ItemEnum) -> data::EnumTy {
    data::EnumTy {
        name: item.ident.to_string(),
        values: item.variants.iter().enumerate().map(|(i, v) | (v.ident.to_string(), i)).collect()
    }
}

fn parse_type_string(ty: String) -> data::Types {
    use data::Types::*;
    match ty.as_str() {
        "str" | "String" => String,
        "bool" => Bool,
        "f32" => F32,
        "f64" => F64,
        "i8" => I8,
        "i16" => I16,
        "i32" => I32,
        "i64" => I64,
        "u8" => U8,
        "u16" => U16,
        "u32" => U32,
        "u64" => U64,
        _ => Named(ty)
    }
}

fn parse_type(item: &syn::Type) -> data::Types {
    match item {
        syn::Type::Path(path) => {
            let segments = &path.path.segments;
            if segments.len() != 1 {
                emit_error(vec![(item.span(), "Path segments with len != 1")]);
            }
            let segment = &segments[0];
            if !segment.arguments.is_empty() {
                let args = match &segment.arguments {
                    syn::PathArguments::AngleBracketed(v) => v,
                    _ => emit_error(vec![(item.span(), "Angle bracketed arguments expected")])
                };
                let types = args.args.iter().map(|arg| match arg {
                    syn::GenericArgument::Type(v) => parse_type(v),
                    _ => emit_error(vec![(item.span(), "Only types are supported")])
                }).collect::<Vec<_>>();
                let name = segment.ident.to_string();
                if name == "Option" {
                    if types.len() != 1 { emit_error(Some((segment.span(), "Option needs exactly one argument"))); }
                    data::Types::Optional(types[0].clone().into())
                } else if name == "Vec" {
                    if types.len() != 1 { emit_error(Some((segment.span(), "Vec needs exactly one argument"))); }
                    data::Types::Array(types[0].clone().into())
                } else {
                    data::Types::Generic(name, types)
                }
            } else {
                parse_type_string(segment.ident.to_string())
            }
        }
        syn::Type::Slice(slice) => {
            data::Types::Array(parse_type(&slice.elem).into())
        }
        _ => emit_error(vec![(item.span(), "Unsupported type")])
    }
}

fn parse_struct(item: &syn::ItemStruct) -> data::StructTy {
    let name = item.ident.to_string();
    if let Some(v) = &item.generics.where_clause {
        emit_error(Some((v.span(), "Where clauses are not allowed")));
    }
    if item.generics.params.len() > 1 {
        emit_error(Some((item.generics.params.span(), "Only one generic parameter is allowed for now")));
    }
    let generic_names = item.generics.params.iter().map(|g| {
        match g {
            syn::GenericParam::Const(_) |
            syn::GenericParam::Lifetime(_) => emit_error(Some((g.span(), "Only generic types are allowed"))),
            syn::GenericParam::Type(ty) => {
                if !ty.bounds.is_empty() {
                    emit_error(Some((ty.span(), "Bounds are not allowed")));
                }
                if let Some(d) = &ty.default {
                    emit_error(Some((d.span(), "Defaults are not allowed")));
                }
                ty.ident.to_string()
            }
        }
    }).collect();
    let fields = item.fields.iter().map(|field| {
        if field.ident.is_none() {
            emit_error(vec![(field.span(), "Missing field name")]);
        }
        let name = field.ident.as_ref().unwrap().to_string();
        let ty = parse_type(&field.ty);
        data::FieldTy { name, ty }
    }).collect();
    data::StructTy { name, fields, generic_names }
}

fn try_parse_iterator(ty: &syn::Type) -> Option<data::Types> {
    if let syn::Type::Path(ty) = ty {
        let seg = ty.path.segments.last()?;
        if seg.ident.to_string() == "Iterator" {
            if let syn::PathArguments::AngleBracketed(args) = &seg.arguments {
                if let Some(syn::GenericArgument::Type(ty)) = args.args.first() {
                    Some(parse_type(ty))
                } else { None }
            } else { None }
        } else { None }
    } else { None }
}

fn parse_method(item: &syn::Signature) -> data::MethodTy {
    let mut method = data::MethodTy::default();
    method.name = item.ident.to_string();

    for arg in &item.inputs {
        let arg = match arg {
            syn::FnArg::Typed(v) => v,
            _ => emit_error(vec![(arg.span(), "Unsupported argument")])
        };
        let ty = parse_type(&arg.ty);
        let name = match &*arg.pat {
            syn::Pat::Ident(v) => v.ident.to_string(),
            _ => emit_error(vec![(arg.span(), "Unsupported argument")])
        };
        method.args.push(data::FieldTy { name, ty });
    }

    match &item.output {
        syn::ReturnType::Default => {
            method.ret = None;
            method.ret_stream = false;
        }
        syn::ReturnType::Type(_, ty) => {
            if let Some(ty) = try_parse_iterator(ty) {
                method.ret_stream = true;
                method.ret = Some(ty);
            } else {
                method.ret_stream = false;
                method.ret = Some(parse_type(ty));
            }
        }
    };

    method
}

fn parse_service(item: &syn::ItemTrait) -> data::ServiceTy {
    let name = item.ident.to_string();
    let mut methods = vec![];
    for item in &item.items {
        let item = match item {
            syn::TraitItem::Fn(v) => v,
            _ => emit_error(vec![(item.span(), "Only functions are supported")])
        };
        methods.push(parse_method(&item.sig));
    }
    data::ServiceTy { name, methods }
}

pub fn parse_file(ast: &syn::File) -> data::RPC {
    let mut rpc = data::RPC::default();

    for item in &ast.items {
        match item {
            syn::Item::Enum(v) => rpc.enums.push(parse_enum(v)),
            syn::Item::Struct(v) => rpc.structs.push(parse_struct(v)),
            syn::Item::Trait(v) => rpc.services.push(parse_service(v)),
            syn::Item::Use(_) => {}
            _ => emit_error(vec![(item.span(), "Unsupported item")])
        }
    }

    rpc
}
